//------------------------------------------------------------------------------
// Copyright (C) 2011, Robert Johansson, Raditex AB
// All rights reserved.
//
// FreeSCADA 
// http://www.FreeSCADA.com
// freescada@freescada.com
//
//------------------------------------------------------------------------------

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <syslog.h>
#include <string.h>

#include <stdio.h>
#include <mbus.h>

#include "mbus-tcp.h"

static mbus_tcp_handle *handle = NULL;
static int debug = 0;

//------------------------------------------------------------------------------
// send a data request packet to from master to slave: the packet selects
// a slave to be the active secondary addressed slave if the secondary address
// matches that of the slave.
//------------------------------------------------------------------------------
int
send_select_frame(mbus_tcp_handle *handle, char *secondary_addr_str)
{
    int val;
    mbus_frame *frame;
        
    frame = mbus_frame_new(MBUS_FRAME_TYPE_LONG);

    if (mbus_frame_select_secondary_pack(frame, secondary_addr_str) == -1)
    {
        fprintf(stderr, "%s: failed to pack selection mbus frame\n", __PRETTY_FUNCTION__);
        mbus_frame_free(frame);
        return -1;
    }    

    if (mbus_tcp_send_frame(handle, frame) == -1)
    {
        fprintf(stderr, "%s: failed to send mbus frame\n", __PRETTY_FUNCTION__);
        mbus_frame_free(frame);
        return -1;
    }    
	
    return 0;
}


//------------------------------------------------------------------------------
// send a data request packet to from master to slave
//------------------------------------------------------------------------------
int
send_data_request_frame(mbus_tcp_handle *handle, int address)
{
    int retval = 0;
    mbus_frame *frame;

    frame = mbus_frame_new(MBUS_FRAME_TYPE_SHORT);
    frame->control  = MBUS_CONTROL_MASK_REQ_UD2 | MBUS_CONTROL_MASK_DIR_M2S; 
    frame->address  = address;

    if (mbus_tcp_send_frame(handle, frame) == -1)
    {
        fprintf(stderr, "%s: failed to send mbus frame\n", __PRETTY_FUNCTION__);
        retval = -1;
    }

    mbus_frame_free(frame);
    return retval;
}


//------------------------------------------------------------------------------
// Check if there are any devices matching the given address mask.
//------------------------------------------------------------------------------
#define MBUS_PROBE_NOTHING   0
#define MBUS_PROBE_SINGLE    1
#define MBUS_PROBE_COLLISION 2
#define MBUS_PROBE_ERROR     -1

static long probe_count = 0;
static int  match_count = 0;

int
mbus_probe_secondary_address(char *mask, char *matching_addr)
{
    int ret;
    mbus_frame reply;

    if (mask == NULL || matching_addr == NULL || strlen(mask) != 16)
    {
        fprintf(stderr, "%s: Invalid address masks.\n", __PRETTY_FUNCTION__);
        return MBUS_PROBE_ERROR;
    }

    probe_count++;

    // send select command
    if (send_select_frame(handle, mask) == -1)
    {
        fprintf(stderr, "%s: Failed to send selection frame: %s\n", __PRETTY_FUNCTION__, mbus_error_str());
        return MBUS_PROBE_ERROR;
    }

    ret = mbus_tcp_recv_frame(handle, &reply);

    if (ret == -1)
    {
        return MBUS_PROBE_NOTHING;
    }    

    if (ret == -2)
    {
        return MBUS_PROBE_COLLISION;
    }

    if (mbus_frame_type(&reply) == MBUS_FRAME_TYPE_ACK)
    {
        // send a data request command to find out the full address
        if (send_data_request_frame(handle, 253) == -1)
	    {
            fprintf(stderr, "%s: Failed to send request to selected secondary device [mask %s]: %s\n", __PRETTY_FUNCTION__, mask, mbus_error_str());
            return MBUS_PROBE_ERROR;
        }

        if (mbus_tcp_recv_frame(handle, &reply) == -1)
        {
            return MBUS_PROBE_NOTHING;
        }

        if (mbus_frame_type(&reply) != MBUS_FRAME_TYPE_ACK)
        {
            snprintf(matching_addr, 16, "%s", mbus_frame_get_secondary_address(&reply));
            match_count++;
            return MBUS_PROBE_SINGLE;
        }
        else
        {
            fprintf(stderr, "%s: Error: Unexpected reply for address [%s]. Got ACK, expected data.\n", __PRETTY_FUNCTION__, mask);
            return MBUS_PROBE_NOTHING;
        }
    }

    fprintf(stderr, "%s: Error: Unexpected reply for address [%s]\n", __PRETTY_FUNCTION__, mask);
    return MBUS_PROBE_NOTHING;
}

//------------------------------------------------------------------------------
// Execution starts here:
//------------------------------------------------------------------------------
int
main(int argc, char **argv)
{
    mbus_frame reply;
    mbus_frame_data reply_data;

    char *host, matching_addr[16];
    int port, address;

    if (argc != 4)
    {
        fprintf(stderr, "usage: %s host port mbus-address\n", argv[0]);
        return 0;
    }
 
    host = argv[1];   
    port = atoi(argv[2]);
    
    if ((handle = mbus_tcp_connect(host, port)) == NULL)
    {
        fprintf(stderr, "Failed to setup connection to M-bus gateway\n");
        return -1;
    }

    if (strlen(argv[3]) == 16)
    {
        // secondary addressing

        int probe_ret;

        probe_ret = mbus_probe_secondary_address(argv[3], matching_addr);

        if (probe_ret == MBUS_PROBE_COLLISION)
        {
            fprintf(stderr, "%s: Error: The address mask [%s] matches more than one device.\n", __PRETTY_FUNCTION__, argv[3]);
            return -1;
        }
        else if (probe_ret == MBUS_PROBE_NOTHING)
        {
            fprintf(stderr, "%s: Error: The selected secondary address does not match any device [%s].\n", __PRETTY_FUNCTION__, argv[3]);
            return -1;
        }
        else if (probe_ret == MBUS_PROBE_ERROR)
        {
            fprintf(stderr, "%s: Error: Failed to probe secondary address [%s].\n", __PRETTY_FUNCTION__, argv[3]);
            return -1;    
        }
        // else MBUS_PROBE_SINGLE

        if (send_data_request_frame(handle, 253) == -1)
        {
            fprintf(stderr, "Failed to send M-Bus request frame.\n");
            return -1;
        }
    } 
    else
    {
        // primary addressing

        address = atoi(argv[3]);
        if (send_data_request_frame(handle, address) == -1)
        {
            fprintf(stderr, "Failed to send M-Bus request frame.\n");
            return -1;
        }
    }   

    if (mbus_tcp_recv_frame(handle, &reply) == -1)
    {
        fprintf(stderr, "Failed to receive M-Bus response frame.\n");
        return -1;
    }    

    //
    // parse data and print in XML format
    //
    if (debug)
    {
        mbus_frame_print(&reply);
    }
    mbus_frame_data_parse(&reply, &reply_data);
    printf("%s", mbus_frame_data_xml(&reply_data));

    mbus_tcp_disconnect(handle);
    return 0;
}



