//------------------------------------------------------------------------------
// Copyright (C) 2011, Robert Johansson, Raditex AB
// All rights reserved.
//
// FreeSCADA 
// http://www.FreeSCADA.com
// freescada@freescada.com
//
//------------------------------------------------------------------------------

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <syslog.h>
#include <string.h>

#include <stdio.h>
#include <mbus.h>

#include "mbus-tcp.h"

static mbus_tcp_handle *handle = NULL;


//------------------------------------------------------------------------------
// send a data request packet to from master to slave: the packet selects
// a slave to be the active secondary addressed slave if the secondary address
// matches that of the slave.
//------------------------------------------------------------------------------
int
send_select_frame(mbus_tcp_handle *handle, char *secondary_addr_str)
{
    int val;
    mbus_frame *frame;
        
    frame = mbus_frame_new(MBUS_FRAME_TYPE_LONG);

    if (mbus_frame_select_secondary_pack(frame, secondary_addr_str) == -1)
    {
        fprintf(stderr, "%s: failed to pack selection mbus frame\n", __PRETTY_FUNCTION__);
        mbus_frame_free(frame);
        return -1;
    }    

    if (mbus_tcp_send_frame(handle, frame) == -1)
    {
        fprintf(stderr, "%s: failed to send mbus frame\n", __PRETTY_FUNCTION__);
        mbus_frame_free(frame);
        return -1;
    }    
	
    return 0;
}


//------------------------------------------------------------------------------
// send a data request packet to from master to slave
//------------------------------------------------------------------------------
int
send_data_request_frame(mbus_tcp_handle *handle, int address)
{
    int retval = 0;
    mbus_frame *frame;

    frame = mbus_frame_new(MBUS_FRAME_TYPE_SHORT);
    frame->control  = MBUS_CONTROL_MASK_REQ_UD2 | MBUS_CONTROL_MASK_DIR_M2S; 
    frame->address  = address;

    if (mbus_tcp_send_frame(handle, frame) == -1)
    {
        fprintf(stderr, "%s: failed to send mbus frame\n", __PRETTY_FUNCTION__);
        retval = -1;
    }

    mbus_frame_free(frame);
    return retval;
}


//------------------------------------------------------------------------------
// Check if there are any devices matching the given address mask.
//------------------------------------------------------------------------------
#define MBUS_PROBE_NOTHING   0
#define MBUS_PROBE_SINGLE    1
#define MBUS_PROBE_COLLISION 2
#define MBUS_PROBE_ERROR     -1

static long probe_count = 0;
static int  match_count = 0;

int
mbus_probe_secondary_address(char *mask, char *matching_addr)
{
    int ret;
    mbus_frame reply;

    if (mask == NULL || matching_addr == NULL || strlen(mask) != 16)
    {
        fprintf(stderr, "%s: Invalid address masks.\n", __PRETTY_FUNCTION__);
        return MBUS_PROBE_ERROR;
    }

    probe_count++;

    // send select command
    if (send_select_frame(handle, mask) == -1)
    {
        fprintf(stderr, "%s: Failed to send selection frame: %s\n", __PRETTY_FUNCTION__, mbus_error_str());
        return MBUS_PROBE_ERROR;
    }

    ret = mbus_tcp_recv_frame(handle, &reply);

    if (ret == -1)
    {
        return MBUS_PROBE_NOTHING;
    }    

    if (ret == -2)
    {
        return MBUS_PROBE_COLLISION;
    }

    if (mbus_frame_type(&reply) == MBUS_FRAME_TYPE_ACK)
    {
        // send a data request command to find out the full address
        if (send_data_request_frame(handle, 253) == -1)
	    {
            fprintf(stderr, "%s: Failed to send request to selected secondary device [mask %s]: %s\n", __PRETTY_FUNCTION__, mask, mbus_error_str());
            return MBUS_PROBE_ERROR;
        }

        if (mbus_tcp_recv_frame(handle, &reply) == -1)
        {
            return MBUS_PROBE_NOTHING;
        }

        if (mbus_frame_type(&reply) != MBUS_FRAME_TYPE_ACK)
        {
            snprintf(matching_addr, 17, "%s", mbus_frame_get_secondary_address(&reply));
            match_count++;
            return MBUS_PROBE_SINGLE;
        }
        else
        {
            fprintf(stderr, "%s: Error: Unexpected reply for address [%s]. Got ACK, expected data.\n", __PRETTY_FUNCTION__, mask);
            return MBUS_PROBE_NOTHING;
        }
    }

    fprintf(stderr, "%s: Error: Unexpected reply for address [%s]\n", __PRETTY_FUNCTION__, mask);
    return MBUS_PROBE_NOTHING;
}

//------------------------------------------------------------------------------
// Iterate over all address masks according to the M-Bus probe algorithm.
//------------------------------------------------------------------------------
int
scan_2nd_address_range(int pos, char *addr_mask)
{
    int i, i_start, i_end, probe_ret;
    char *mask, matching_mask[17];

    if (strlen(addr_mask) != 16)
    {
        fprintf(stderr, "%s: Illegal address mask [%s]. Not 16 characters long.\n", __PRETTY_FUNCTION__, addr_mask);
        return -1;
    }

    if (pos < 0 || pos >= 16)
    {
        return 0;
    }

    if ((mask = strdup(addr_mask)) == NULL)
    {
        fprintf(stderr, "%s: Failed to allocate local copy of the address mask.\n", __PRETTY_FUNCTION__);
        return -1;
    }
    
    if (mask[pos] == 'f' || mask[pos] == 'F')
    {
        i_start = 0;
        i_end   = 9;
    }
    else
    {
        if (pos < 15)
        {
            scan_2nd_address_range(pos+1, mask);
        }
        else
        {
            i_start = (int)(mask[pos] - '0');
            i_end   = (int)(mask[pos] - '0');
        }
    }

    for (i = 0; i <= 9; i++)
    {
        mask[pos] = '0'+i;
    
        probe_ret = mbus_probe_secondary_address(mask, matching_mask);

        if (probe_ret == MBUS_PROBE_SINGLE)
        {
            printf("Found a device on secondary address %s [using address mask %s]\n", matching_mask, mask);
        }
        else if (probe_ret == MBUS_PROBE_COLLISION)
        {
            // collision, more than one device matching, restrict the search mask further
            scan_2nd_address_range(pos+1, mask);
        }
        else if (probe_ret == MBUS_PROBE_NOTHING)
        {
             // nothing... move on to next address mask
        }
        else // MBUS_PROBE_ERROR
        {
            fprintf(stderr, "%s: Failed to probe secondary address [%s].\n", __PRETTY_FUNCTION__, mask);
            return -1;    
        }
    }

    free(mask);
    return 0;
}

//------------------------------------------------------------------------------
// Execution starts here:
//------------------------------------------------------------------------------
int
main(int argc, char **argv)
{
    char *host, *addr_mask;
    int port, address;

    if (argc != 4 && argc != 3)
    {
        fprintf(stderr, "usage: %s host port [address-mask]\n", argv[0]);
        fprintf(stderr, "\trestrict the search by supplying an optional address mask on the form\n");
        fprintf(stderr, "\t'FFFFFFFFFFFFFFFF' where F is a wildcard character\n");
        return 0;
    }
 
    host = argv[1];   
    port = atoi(argv[2]);
    if (argc == 4)
    {
        addr_mask = strdup(argv[3]);
    }
    else
    {
        addr_mask = strdup("FFFFFFFFFFFFFFFF");
    }

    if (strlen(addr_mask) != 16)
    {
        fprintf(stderr, "Misformatted secondary address mask. Must be 16 character HEX number.\n");
        return -1;
    }

    if ((handle = mbus_tcp_connect(host, port)) == NULL)
    {
        fprintf(stderr, "Failed to setup connection to M-bus gateway: %s\n", mbus_error_str());
        return -1;
    }

    scan_2nd_address_range(0, addr_mask);

    mbus_tcp_disconnect(handle);

    printf("Summary: Tried %ld address masks and found %d devices.\n", probe_count, match_count);

    return 0;
}

